import java.awt.Point;
import java.awt.Image;
import java.io.*;
import java.net.*;
import Map;
import ClientThread;
import ClientFrame;

// ********************************************************************************
//			ClientGame
// ********************************************************************************

// Fornisce un interfaccia fra la classe che gestisce la comunicazione con 
// il Giocatore e la classe che gestisce la comunicazione con Server. 
// Un'azione del giocatore o una cominicazione del Server provocano una 
// chiamata a questa classe.
// Questa classe fornisce quindi le primitive che eseguono i comandi del 
// Giocatore e del Server.
// Tali primitive sono ad acesso mutuamente esclusivo ( modificatore sychronized)
// per evitare che contemporaneamente vengano gestiti piu' comandi.
// Inotre dato che i vari metodi terminano sempre ( non ci sono loop ) e vi e'
// un unico semaforo a livello di classe non c'e' possibilita di blocco critico.


public class ClientGame extends Map {


  private int 		player;
  private String	name;

	// Numero e nome del giocatore

  private boolean	check;

	// Indica il livello di controllo effettuato dal client sulla valitita' 
	// delle mosse fatte dal giocatore
	//   false : Nessun controllo, tutti i controlli vengono lasciati al
	//	     server.	
	//   true  : Il Client invia al Server solo mosse valide

  private ClientThread	thread;

	// Classe per la gestione dell'IO sul Socket

  private ClientFrame	frame;

	// Classe per la gestione della grafica e della cominicazione con 
	// il giocatore



  public ClientGame( String name,   boolean check,
                     String Server, int Port, 
				    Image Img[]    ) throws IOException {

	// Costruttore
	//   name	: nome del giocatore
	//   check	: livello di controllo sulla validita' delle mosse
	//   Server	: URL Server
	//   Port	: numero di porta
	//   Img[]	: Immagini dei pezzi

    super( new Point(14,14) );		// Costruttore classe Map
    this.name = new String ( name );	// Copia locale delle variabili
    this.check = check; 

	// Crea Socket con il Server

    Socket socket = new Socket(Server, Port);  

	// Crea il Thread per la gestione dell'io sul Socket
    
    thread = new ClientThread( socket , this ); 
	
	// Crea la Classe per la gestione della grafica e della comunicazione
	// con il giocatore

    frame = new ClientFrame( Img, this );

	// Attiva il Thread

    new Thread( thread ).start(); 
 }


  public synchronized int getPlayer() {

	// Restituisce il numero di giocatore

   return player;
  }


// I metodi do* vengono chiamati dalla classe per la gestione della comunicazione
// con il Server, ( comunicazione in ingresso dal Server ) e' definito un metodo
// per ogni comando trasmesso dal Server.
// I metodi sono ad acesso mutuamente escusivo in modo che venga gestito un
// comando alla volta.
// 
// Il socket puo' essere chiuso ( ad asempio nel caso di abbandono ) con il
// buffer di ingresso non vuoto. La classe per la comunicazione con il Server
// legge in ogni caso i comandi presenti in questo buffer ed effettua una chiamata 
// ad un metodo do*. Affinche questi comandi non vengano eseguiti, insieme alla 
// chiusura del socket, 'thread' ( istanza della classe ClientThread ) viene imposto 
// a null e viene aggiunta come prima riga di ogni metodo do* :
//   if ( thread == null ) return;
// Nel caso che il socket venga chiuso 'thread==null' e il comando non viene
// gestito
//
// Se nella comunicazione con il Server viene generato un IOException viene richiamato
// doDEAD() che termina la partita.
//


// In questi metodi :
//
//   frame.appErr(string)	effettua output nella finesra di testo Error
//   frame.appQUOTE(string) 	effettua output nella finesra di testo Quote
//   frame.setLabel(int,string)	definisce una delle label riservate ai nomi	
//   frame.cleanLabel(int)	elimina una label e rincompatta le sucessive   

  public synchronized void doHELO() {

	// Gestisce il comando HELO, al quale il Client risponde comunicando
	// in proprio nome. Nel caso di IOException la partita viene terminata

    if ( thread == null ) return;
    frame.appErr("C <- S : HELO  \n");     
    frame.appErr("C -> S : NAME "+name+" \n");
    try {
      thread.writeNAME(name); 
    } catch ( IOException e ) {
      doDEAD(e);
    }
  }


  public synchronized void doNAME( int player, String name) {

	// Gestisce il comando NAME, viene comunicato al giocatore il nome
	// del giocatore 'player' stampando a video player+" "+name

    if ( thread == null ) return;
    frame.appErr("C <- S : NAME "+player+","+name+" \n");     
    frame.setLabel(player, " -"+player+"- "+name);
  }


  public synchronized void doCLEAN( int player) {

	// Gestisce il comando CLEAN, viene cancellato il giocatore specificato e 
	// ricompattati i sucessivi.

    if ( thread == null ) return;
    frame.appErr("C <- S : CLEAN "+player+" \n");     
    frame.cleanLabel(player);
  }


  public synchronized void doPLAY( int player , int numplayer) {

	// Gestisce il comando PLAY, viene impostato il numero totale 
	// di giocatori, e il numero di giocatore assegnato a questo Client

    if ( thread == null ) return;
    frame.appErr("C <- S : PLAY "+player+","+numplayer+" \n");
    setNumPlayer(numplayer);        
    this.player = player;  
  }


  public synchronized void doBLOCK( int blk1, int blk2) {

	// Gestisce il comando BLOCK, viene aggiunto al giocatore il pezzo
	// specificato

    if ( thread == null ) return;
    frame.appErr("C <- S : BLOCK "+blk1+","+blk2+" \n");     
    if ( putBlock( new Point( blk1,blk2),  player) != 0 )
	// Si verifica quando il Server trasmette due volte lo stesso pezzo,
	// puo' avvenire solo nel caso di codice non corretto....fine partita 
      doDEAD( new Exception( "Pezzo comunicato dal Server errato "));
  }


  public synchronized void doSTART() {

	// Gestisce il comando START, viene dato inizio alla partita

    if ( thread == null ) return;
    frame.appErr("C <- S : START \n"); 
    startPlay();
    frame.createTable();
  }


  public synchronized void doMOVE( int player, int blk1, int blk2, int x, int y, int mode ) {

	// Gestisce il comando MOVE, viene eseguita la mossa specificata  

    if ( thread == null ) return;
    frame.appErr("C <- S : MOVE "+player+","+blk1+","+blk2+","+x+","+y+","+mode+" \n");

	// Se non ha mosso il giocatore associato a questo client la mossa viene
	// data al giocatore fittizio ( player = 0 ) che rappresenta l'insieme
	// degli avversari
    if ( player != this.player ) player = 0;
    int err = putMap( player, new Point(blk1,blk2), new Point(x,y),  mode ); 

    if ( err == 0 ) frame.paint();	// Provoca l'aggiornamento dello schermo
    else 
	// Si verifica quando il Server ha trasmesso una mossa non valida,
	// puo' avvenire solo nel caso in cui il codice sia non corretto
	// La partita viene terminata
      doDEAD(new Exception("Mossa comunicata dal Server non valida"));
  }


  public void doGO() {

	// Gestisce il comando GO, tocca al giocatore associato a questo Client
	// muovere, viene abilitato il pulsante MOVE 

    if ( thread == null ) return;
    frame.appErr("C <- S : GO  \n");
    frame.setEnabledMove( true );	// Abilita il pulsante MOVE
   }


  public void doQUOTE( int player, String quote) {

	// Gestisce il comando QUOTE, viene aggiunto il quote specificato
	// alla finestra di testo quote 

    if ( thread == null ) return;
    frame.appErr("C <- S : QUOTE "+player+","+quote+"  \n");
    if ( player == 0 ) frame.outErr(quote);
    else frame.appQuote("Player : "+player+","+quote+" \n");  
  }
  
      
  public synchronized void doERROR( String error) {

	// Gestisce l'evento "comando letto non conosciuto"

    if ( thread == null ) return;
    frame.appErr("C <- S : "+ error+" \n");
  }


  public synchronized void doEND() {

	// Gestisce il comando END

    doDEAD(null);			// Rilascia le risorse 
    frame.setEnabledMove(false);	// Disabilita il pulsante MOVE
    frame.setEnabledQuote(false); 	// Disabilita il pulsante QUOTE
    frame.outErr(" The END ");
  }


  public synchronized void doDEAD( Exception err ) {

	// Fine Partita, rilascia le risorve 

    if ( thread == null ) return;

	// Nel caso in cui "err==null" non viene fatto l'output delle causee
	// che hanno provocato la fine della partita
	// Viene chiamata doDEAD(null) in doEND() e in risposta al comando exit
	// effettuato dal giocatore. I entrambi i casi non e' un errore ad
	// aver provocato la fine della partita ma una richiesta esplicita 
    if ( err != null ) {
      frame.appErr("ERROR :"+err);	// Output causee fine partita
      frame.outErr("ERROR :"+err);
      frame.setEnabledMove(false);	// Disabilita il pulsante MOVE
      frame.setEnabledQuote(false);	// Disabilita il pulsante QUOTE
    }
    thread.disableThread();		// Chiude il Socket
    thread = null;
	// Impostando 'thread = null' si evita di considerare comandi rimasti 
	// nel buffer di ingresso del socket e si perdono i riferimneti 
	// all'istanza della classe ClientThread utilizzata fino ad adesso.
	// L'area di memoria di una classe senza riferimneti viene automaticamente
	// recuperata dal Garbage Collector  
  }


// I metodi write* vengono chiamati dalla classe per la gestione della comunicazione
// con il giocatore, ( comunicazione in ingresso dal giocatore ) e' definito un metodo
// per ogni comando richiesto dal giocatore.
// I metodi sono ad acesso mutuamente escusivo in modo che venga gestito un
// comando alla volta.
// Anche questi metodi, per i motivi spiegati precedentemente, hanno come prima riga :
//    if ( thread == null ) return


  public synchronized void writeSTART() {

	// Gestisce la richiesta di inizio partita da parte del giocatore, viene
	// comunicato il comando START al Server

    if ( thread == null ) return;
    frame.appErr("C -> S : START \n");
    try { 
      thread.writeSTART();
    } catch ( IOException e ) {
      doDEAD(e);
    }
  }

 
  public synchronized void writeMOVE ( int pz1, int pz2, 
 			               int pnt1, int pnt2, int rot ) {

	// Gestisce una mossa da parte del giocatore, la mossa richiesta viene inviata 
	// al Server
	// Se 'check==true' la mossa viene inviata solo se e' corretta ed e'
	// il turno del giocatore associato a questo client. 

    if ( thread == null ) return;
    if ( check ) {	// se '==true' vengono effettuati i controlli sulla correttezza

      int err = check(player, new Point(pz1,pz2), new Point(pnt1,pnt2),rot);
      if ( err < 0 ) {	// Mossa non e' corretta, esce

	frame.outErr("Mossa non valida: "+ errorString(err) );
	return;
			// Atrimenti disabilita pulsante Move, che verra riabilitato
			// all'arrivo di un GO dal Server
      } else frame.setEnabledMove( false );
    }        
    frame.appErr("C -> S : MOVE "+pz1+","+pz2+","+pnt1+","+pnt2+","+rot+" \n");
    try {
      thread.writeMOVE ( pz1,pz2,pnt1,pnt2,rot);	// Invia la mossa al Server
    } catch ( IOException e ) {
      doDEAD(e);
    }
  }


  public synchronized void writeQUOTE( String quote) {

	// Gestisce la richiesta da parte del giocatore di invio di un messaggio.
	// Il messaggio viene comunicato al Server

    if ( thread == null ) return;
    frame.appErr("C -> S : QUOTE "+quote+"  \n");
    try { 
      thread.writeQUOTE( quote );
    } catch ( IOException e ) {
      doDEAD(e);
    }
  }

}  // class      
