import java.awt.Point;
import java.net.*;
import java.io.*;
import ServerGame;
import Parser;
import ServerLog;

// **************************************************************************
//		ServerThread (thread)
// **************************************************************************

// Thread che gestisce tutte le comunicazioni con gli eventuali clients
// connessi.
//  composto da un thread (metodo run()) che esegue il parsing dei comandi
// di ingresso, chiamando le funzioni do* synchro di ServerGame e una serie
// di metodi write* che scrivono sul socket; questi non serve che siano
// sincronizzati, visto che vengono richiamati dalle do*, che sono synchro,
// di ServerGame
//
// In breve c' un'unica sincronizzazione, a livello di ServerGame.


class ServerThread  implements Runnable {

	// Puntatore alla classe ServerGame
  private ServerGame	game;

	// Interprete comandi in ingresso dal Socket
  Parser 		parser;

	// numero di questo thread, usato come identificativo
  private int 		thread;

	// Variabili interne (socket e streams di IO)
  private Socket 	socket;
  private BufferedReader is;
  private PrintWriter	os;

	// Flag usato per uscire da run()
  private boolean	flag = true;

	// TimeOut di lettura dal socket
  final public static int timeOut = 600000;	// millesimi di secondo



  public ServerThread ( int thread , Socket socket , ServerGame game ) throws IOException {

	// costruttore di ServerParser
    super ();

	// assegnazione delle variabili locali
    this.thread = thread;
    this.socket = socket;
    this.game   = game;

	// Imposta Timeout
    socket.setSoTimeout( timeOut ) ;


	// Crea stream di input ed output
    is = new BufferedReader( 
             new InputStreamReader (
                 new BufferedInputStream (   
                     socket.getInputStream() ) ) );

    os = new PrintWriter( socket.getOutputStream() );


	// Invia HELO al client
    writeHELO();    

	// Definizione interprete di comandi

    int numComm = 4;
    String dataComm[];  
    dataComm = new String[numComm];
 
    dataComm[0] = new String("NAME String");
    dataComm[1] = new String("START");
    dataComm[2] = new String("MOVE Integer,Integer,Integer,Integer,Integer");
    dataComm[3] = new String("QUOTE String");

    parser = new Parser( dataComm , numComm );
  }


	// Questo thread legge dal socket ed esegue il parsing dei comandi
	// che arrivano; se il parsing ha successo chiama una delle funzioni
	// di cambiamento di stato del gioco (do*) di ServerGame. Queste sono
	// synchro, e quindi non ci possono essere race condition.

  public void run() {
    
    String str = null;

    do {
		// inspiegabilmente questa ogni tanto funziona e ogni tanto
		// no...
      try { str = is.readLine(); }
      catch (IOException e) {
        game.doDEAD (thread,e);
        return;
      }   
		// mentre solitamente la chiusura di un socket provoca un
		// ``null'' come stringa di ritorno, prima o poi...
		// o meglio, appena si svuota il buffer!
      if ( str == null ) {
        game.doDEAD (thread,new IOException("socket closed by peer"));
        return;
      }

	// il problema qui  che se il server chiude forzatamente la
	// connessione, e il client stava spedendo qualcosa, questo resta
	// in qualche modo in buffer, fino all'effetivo null.
	// Non genera errori (perch in ServerGame il thread  stato
	// disabilitato, ma da fastidio...

	// faccio quindi un doppio controllo, prima e dopo.
      if ( !flag ) return;

      switch ( parser.setCommand(str) ) {

        case 0 :	// Nome
          game.doNAME ( thread , parser.getString(0) );
          break;
          
        case 1 :	// START
          game.doSTART ( thread );
          break;

        case 2 : 	// Nuova Mossa
          game.doMOVE ( thread , parser.getInteger(0).intValue(), 
			parser.getInteger(1).intValue(), parser.getInteger(2).intValue(), 
			parser.getInteger(3).intValue(), parser.getInteger(4).intValue() );
          break;

        case 3 :	// Nuovo QUOTE
          game.doQUOTE ( thread , parser.getString(0) );
          break;
        
	default :
          ServerLog.doLog(5,game.getNumGame(),thread,"ServerThread.run: parse error: "+str);  
      } 
    	      
    } while ( flag ); 

  }


	// non penso che servano spiegazioni... ;)))

  public void disableThread() {

    ServerLog.doLog(2,game.getNumGame(),thread,"ServerThread.disableThread: socket closed, thread disposed");

    flag = false;

    String command = "BYE";

    os.println( command );
    os.flush();

    try {
      os.close();
      is.close();
      socket.close();    
    } catch (IOException e) {}
  
  }



// ***************************************************************************
//
// 	Metodi write*
//
// Tutti questi metodi sono perfettamente uguali, e di significato
// chiaro, non penso servano ulteriori commenti...


	// Comunica al Client i pezzi assegnati, chiamando una serie di
	// writeBLOCK

  public void writeAllBlocks(int numblock) throws IOException { 

    Point block = new Point(0,0);
    String str;    

    int i,j;

    for ( i=0 ; i<numblock ; i++ ) 
      for ( j=0 ; j<=i ; j++ ) {
       
        block.move(i,j);

        switch ( game.getBlock( thread, block )) { 
	case 0 : 	// Pezzo appartiene al client
	  writeBLOCK ( block.x, block.y );
	  break;
	case -1 :
	  ServerLog.doLog(0,game.getNumGame(),thread,"ServerThread.writeAllBlocks : Errore Ingestibile");
          System.exit(1);
	}
      }
  }


  	// Comando HELO in uscita

  public void writeHELO () throws IOException {
  
    String command = "HELO";

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeHELO : thread=" + thread );
  }


  	// Comando QUOTE in uscita

  public void writeQUOTE ( int play, String quota ) throws IOException {
  
    String command = "QUOTE " + play + "," + quota;

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeQUOTE: IO error");
  }


  	// Comando MOVE in uscita

  public void writeMOVE ( int play, int pz1, int pz2, int pnt1, int pnt2, int rot ) throws IOException {
  
    String command = "MOVE " + play + "," + pz1 + "," + pz2 + "," + pnt1 + "," + pnt2 + "," + rot;

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeMOVE: IO error");
  }


  	// Comando BLOCK in uscita

  public void writeBLOCK ( int pz1, int pz2 ) throws IOException {
  
    String command = "BLOCK " + pz1 + "," + pz2;

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeBLOCk: IO error");
  }


  	// Comando NAME in uscita

  public void writeNAME ( int play , String name ) throws IOException {
  
    String command = "NAME " + play + "," + name;

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeNAME: IO error");
  }


  	// Comando GO in uscita; in aggiunta manda un quote da utente 0 che
  	// causa l'apertura di un popup di avviso.

  public void writeGO () throws IOException {
  
    String command = "GO";

    os.println( command );
    if ( os.checkError() )
      throw new IOException("ServerThread.writeGO: IO error" );
    else {
      os.println( "QUOTE 0,Tocca a te!" );
      if ( os.checkError() )
        throw new IOException("ServerThread.writeGO: IO error" );
    }
      
  }


  	// Comando START in uscita

    public void writeSTART () throws IOException {
  
    String command = "START";

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeSTART: IO error" );
  }

  
  	// Comando PLAY in uscita

    public void writePLAY (int player, int numPlayer) throws IOException {
  
    String command = "PLAY "+ player + "," + numPlayer;

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writePLAY: IO error");
  }


  	// Comando CLEAN in uscita

    public void writeCLEAN ( int player ) throws IOException {
  
    String command = "CLEAN "+ player;

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeCLEAN: IO error");
  }


  	// Comando END in uscita

  public void writeEND () throws IOException {
  
    String command = "END";

    os.println( command );
    if ( os.checkError() ) throw new IOException("ServerThread.writeEND: IO error");
  }

} // ServerThread
